/* -*-C-*-
*******************************************************************************
*
* File:         common.h
* RCS:          $Header: $
* Description:  Common include file for the PPU and the SPUs
* Author:       Fabrizio Petrini        
* Created:
* Modified:
* Language:     C
* Package:      N/A
* Status:       Experimental (Do Not Distribute)
*
*******************************************************************************
*/


#ifndef __common__
#define __common__


#include <inttypes.h>

#define BYTES_PER_WORD      sizeof(unsigned)
#define WORDS_PER_QUADWORD  4
#define BITS_PER_BYTE       8
#define DMA_ALIGNMENT       7


/* minimize the functionality and the size of the communication and
   synchronization runtime */
#undef COMMUNICATION_RUNTIME
//#define COMMUNICATION_RUNTIME

/* remove printfs */
#undef REMOVE_PRINTF
//#define REMOVE_PRINTF

#ifdef REMOVE_PRINTF
#define spe_printf( string, args... ) 
#define spe_fflush( stdout )
#else
#define spe_printf( string, args... ) printf( string, ##args )
#define spe_fflush( stdout ) fflush( stdout )
#endif


/* lightweight profiling of the SPU */
//#undef PROFILE_SPU
#define PROFILE_SPU

/* when set, places performance critical variables into registers */
//#undef USE_ASM_REGISTERS
#define USE_ASM_REGISTERS


#ifdef USE_ASM_REGISTERS
#define REGISTER  register
#define ASM_REGISTER( a )   asm( #a )
#else
#define REGISTER 
#define ASM_REGISTER( a ) 
#endif 

#undef DEBUG
//#define DEBUG

#ifdef DEBUG
#define DB(x)  x
#define UDB(x)
#else
#define DB(x)
#define UDB(x) x
#endif

#undef DEBUG_EVENT
//#define DEBUG_EVENT

#ifdef DEBUG_EVENT
#define DB_EV(x)  x
#define UDB_EV(x)
#else
#define DB_EV(x)
#define UDB_EV(x) x
#endif


#undef DEBUG_DMA
//#define DEBUG_DMA

#ifdef DEBUG_DMA
#define DB_DMA(x)  x
#define UDB_DMA(x)
#else
#define DB_DMA(x)
#define UDB_DMA(x) x
#endif


#undef DEBUG_PTP
//#define DEBUG_PTP

#ifdef DEBUG_PTP
#define DB_PTP(x)  x
#define UDB_PTP(x)
#else
#define DB_PTP(x)
#define UDB_PTP(x) x
#endif


#undef DEBUG_REDUCE
//#define DEBUG_REDUCE

#ifdef DEBUG_REDUCE
#define DB_RED(x)  x
#define UDB_RED(x)
#else
#define DB_RED(x)
#define UDB_RED(x) x
#endif


#undef DEBUG_SYNC_BUG
//#define DEBUG_SYNC_BUG

#ifdef DEBUG_SYNC_BUG
#define DB_SYNC(x)  x
#define UDB_SYNC(x)
#else
#define DB_SYNC(x)
#define UDB_SYNC(x) x
#endif

/* boolean constants */
#define TRUE 1
#define FALSE 0

#define sw_max( a, b ) ( ( (a) > (b) ) ? (a) : (b) )

#define align4    __attribute__ ((aligned (4)))
#define align16   __attribute__ ((aligned (16)))
#define align128  __attribute__ ((aligned (128)))
#define align4096 __attribute__ ((aligned (4096)))
#define unused    __attribute__ ((unused))

/* the decrementer on MAMBO is 1.5 nsec per tick (empirical estimation) */
#define NSEC_PER_TICK_MAMBO 1.5

/* the decrementer tick at 3.2 Ghz is 69.8422 nsec */
#define NSEC_PER_TICK_3_2GHZ  69.8422

/* the decrementer tick at 3.2 Ghz is 69.8422 nsec */
#define NSEC_PER_TICK_2_4GHZ  69.8422

/* 37.5 nsec on the new blades */
#define NSEC_PER_TICK_NEW_BLADES_3_2GHZ   37.5

/* update base on the decrementer resolution */
#define NSEC_PER_TICK NSEC_PER_TICK_NEW_BLADES_3_2GHZ

#define MAX_STRING 128
#define MAX_FILE_NAME 32

/* the maximum number of SPUS: we can connect the two Cell processors
   in a blade in master-slave mode, virtualizing a single system with
   up to 16 SPUs */
#define MAX_SPU 16

/* the current implementation of the Cell processor can send messages
   up to 16K */
#define MAX_DMA_SIZE (1<<14)

/* the local store (LS) size, in bytes */
#define LS_SIZE (1<<18)

/* small and large buffers in main memory; these defines identify the
   number of unsigned int entries in each array, so the actual size of
   the array can be computed by multiplying the define by
   sizeof( unsigned int )  */
#define SMALL_BUFFER_SIZE  (1<<10)
#define LARGE_BUFFER_SIZE  (1<<20)

/* and in the SPE's local storage; it is very important to minimize
   the size of these arrays, when the application is memory
   constrained */
#define SPU_SMALL_BUFFER_SIZE  (1<<4)
#define SPU_LARGE_BUFFER_SIZE  (1<<4)

/* control block size: need to compute this manually to double check
   that ( sizeof( spe_control_block ) == CB_SIZE ); sometimes internal
   alignment problems (for example, unsigned long long ints inside the
   spe_control_block) can make these two values distinct, making life
   very interesting */
#define CB_SIZE 1024

/* the maximum number of monitored events: we monitor events using 32
   bit unsigned integers, using the most significative part of the bit
   string to tag the event which, in most cases identifies a latency
   expressed in decrementer ticks; for example, if we use 5 bits event
   tags, we can monitor 32 different types of events, and we can
   measure intervals up to 2^27 decrementer ticks, which is
   approximately 5.3 seconds at a resolution of 40 nanoseconds per
   decrementer tick; when we overflow the bit threshold, we use three
   words, one to identify an overflow, one for the tag and the third
   one for the actual value, that can use the full 32 bits */
#define MAX_EVENT 32
#define EVENT_TAG_BITS  5
#define EVENT_VALUE_BITS 27
#define EVENT_VALUE_BITMASK  ((1<<EVENT_VALUE_BITS) - 1)

/* specific events: each event should be matched by an entry in the
   dist_name array */
#define RUN_TIME_EVENT              ( 0<<EVENT_VALUE_BITS)
#define ITERATIONS_EVENT            ( 1<<EVENT_VALUE_BITS)
#define DMAS_EVENT                  ( 2<<EVENT_VALUE_BITS)
#define INPUT_DMAS_EVENT            ( 3<<EVENT_VALUE_BITS)
#define OUTPUT_DMAS_EVENT           ( 4<<EVENT_VALUE_BITS)
#define ACTION5_EVENT               ( 5<<EVENT_VALUE_BITS)
#define ACTION6_EVENT               ( 6<<EVENT_VALUE_BITS)
#define ACTION7_EVENT               ( 7<<EVENT_VALUE_BITS)
#define REDUCE_BQ_EVENT             ( 8<<EVENT_VALUE_BITS)
#define REDUCE_Q_EVENT              ( 9<<EVENT_VALUE_BITS)
#define REDUCE_L_EVENT              (10<<EVENT_VALUE_BITS)
#define LOOP_BQ_EVENT               (11<<EVENT_VALUE_BITS)
#define LOOP_Q_EVENT                (12<<EVENT_VALUE_BITS)
#define OVERFLOW_EVENT              (13<<EVENT_VALUE_BITS)



#define distarray							\
  const dist_element dist_array[MAX_EVENT] = {				\
    { .event_id = RUN_TIME_EVENT,    .event_file_name = "run_time" },	\
    { .event_id = ITERATIONS_EVENT,  .event_file_name = "iterations" },	\
    { .event_id = DMAS_EVENT,        .event_file_name = "dmas" },	\
    { .event_id = INPUT_DMAS_EVENT,  .event_file_name = "input_dmas" },	\
    { .event_id = OUTPUT_DMAS_EVENT, .event_file_name = "output_dmas" }, \
    { .event_id = ACTION5_EVENT,     .event_file_name = "action5" },	\
    { .event_id = ACTION6_EVENT,     .event_file_name = "action6" },	\
    { .event_id = ACTION7_EVENT,     .event_file_name = "action7" },	\
    { .event_id = REDUCE_BQ_EVENT,   .event_file_name = "reduce_bQ" },	\
    { .event_id = REDUCE_Q_EVENT,    .event_file_name = "reduce_Q" },	\
    { .event_id = REDUCE_L_EVENT,    .event_file_name = "reduce_L" },	\
    { .event_id = LOOP_BQ_EVENT,     .event_file_name = "loop_bQ" },	\
    { .event_id = LOOP_Q_EVENT,      .event_file_name = "loop_Q" },	\
    { .event_id = OVERFLOW_EVENT,    .event_file_name = "overflow" },	\
  };

/* the size in bytes of a quadword, in bytes */
#define QUADWORD_SIZE     16


/*
*******************************************************************************
*
*                               typedefs
*
*******************************************************************************
*/


/* boolean values */
typedef enum tag_boolean { False, True } boolean_t;

/* effective address */
typedef uint32_t effective_address_t;

/* dma-list */
typedef struct _dma_list_element
{
  union {
    unsigned int all32;
    struct {
      unsigned nbytes: 31;
      unsigned stall:   1;
    } bits;
  } size;
  unsigned int ea_low;
} dma_list_element;


/* to access the 64-bit effective address at the granularity of 32 bit
   words; the most significant (high) word is ui[0], while the low one
   is u[1] */
typedef union
{
  unsigned long long ull;
  unsigned int ui[2];
} my_addr64;


/* communication buffers need to be aligned on a 16-byte boundary, to
   make the DMA engine happy */
typedef union {
  int ival[4];
  unsigned int uival[4];
  float fval[4];
  double dval[2];
  char cval[16];
} communication_buffer;


/* the type of reduce operation */
typedef enum {
  REDUCE_BARRIER,
  REDUCE_INT,
  REDUCE_INT_MAX,
  REDUCE_DOUBLE,
  REDUCE_DOUBLE_MAX
} reduce_op_type;


/* each distribution element is identified by an event id and an event
   filename */
typedef struct _dist_element
{
  const int  event_id;
  const char event_file_name [MAX_FILE_NAME];
} dist_element;


typedef struct _spe_control_block
{
  /* array of pointers to local storages */
  unsigned long long ls[MAX_SPU];

  /* array of pointers to mirrors of local storages, allocated in main
     memory */
  unsigned long long mirror_ls[MAX_SPU];

  /* array of pointers to the SPE problem state control areas */
  unsigned long long ps_cntl[MAX_SPU];

  /* array of pointers to small data buffers, allocated in main memory */
  unsigned long long small_buffer[MAX_SPU];

  /* array of pointers to insertion counters, allocated in main memory */
  unsigned long long in_small_buffer[MAX_SPU];

  /* array of pointers to larger data buffers, allocated in main memory */
  unsigned long long large_buffer[MAX_SPU];

  /* array of pointers to insertion counters, allocated in main memory */
  unsigned long long in_large_buffer[MAX_SPU];

  /* the number of spus */
  unsigned int spus;

  /* the ceiling of the logarithm of the number of SPUs */
  unsigned int log_spus;

  /* pad to a full cache line boundary (128 bytes); we need to be
     careful to proper alignments, the control block must be aligned
     on a 128-byte boundary; */
  char pad[120];
} spe_control_block;


typedef struct user_control_block
{
  /* the number of SPUs */
  unsigned N;
  
  /* the start address of the dynamically allocated buffer, mapped to
     huge pages, where we allocate the stt and the input and output
     buffers */
  unsigned long long dynamic_dma_buffer;
  
  /* the size of the dynamic buffer, in bytes */
  unsigned dynamic_dma_buffer_size;

  /* do not remove this padding field - 
     the align16 keyword keeps the entire structure aligned */
  char filler[1] align16;
} control_block_t;


/* allows multiple access patterns to a quadword */
typedef union {
  unsigned char          bytes[16];    /* as sixteen characters */
  signed char            sbytes[16];   /* as sixteen signed characters */
  unsigned short         shorts[8] ;   /* as eight unsigned shorts */
  signed short           sshorts[8] ;  /* as eight signed shorts */
  unsigned int           words[4];     /* as four unsigned integers */
  signed int             swords[4];    /* as four signed integers */
  float                  floats[4];    /* as four unsigned integers */
  double                 doubles[2];   /* as two doubles */
} quadword_t align16;



#endif /* __common__ */






