/* 
*******************************************************************************
*
* File:         timers.h
* RCS:          $Header: $
* Description:  PPU timers
* Author:       Fabrizio Petrini 
* Created:
* Modified:
* Language:     C
* Package:      N/A
* Status:       Experimental (Do Not Distribute)
*
*******************************************************************************
*/


#ifndef __timers__
#define __timers__


#include "runtime_ppu.h"


/*
*******************************************************************************
*
*                                 defines
*
*******************************************************************************
*/


/* define to use gettimeofday */
#define GETTIME


#define timersub(a, b, result)                                                \
  do {                                                                        \
    (result)->tv_sec = (a)->tv_sec - (b)->tv_sec;                             \
    (result)->tv_usec = (a)->tv_usec - (b)->tv_usec;                          \
    if ((result)->tv_usec < 0) {                                              \
      --(result)->tv_sec;                                                     \
      (result)->tv_usec += 1000000;                                           \
    }                                                                         \
  } while (0)


#define set_time( time, tv_1, tv_2, tv_diff )  \
        timersub(&tv_2, &tv_1, &tv_diff); \
        *time = ((double)(tv_diff.tv_sec) * 1.0E+6 + \
                (double)(tv_diff.tv_usec)); \
        *time = *time / 1.0E+6


/* get_clock set the time based on the timer type, that is a struct
   timeval with gettimeofday
   
   diff( t, t2, t1 ) set the first parameter, which must be a double,
   with the difference between the two timer values, t2 and t1.
   
   the time is returned in seconds
*/
#ifdef GETTIME
#define timer_t struct timeval
#define get_clock(t) gettimeofday( &t, &tz );
#define diff( t, t2, t1 ) { set_time( &t, t1, t2, t_diff ); }
#else
/* some other timer */ 
#endif


/*
*******************************************************************************
*
*                                externals
*
*******************************************************************************
*/


#ifdef GETTIME
/* the timezone, used by gettimeofday */
extern struct timezone tz;
extern struct timeval t_diff;
#endif


/*
*******************************************************************************
*
*                         high resolution timers 
*
*******************************************************************************
*/


#define hires_timer_ull( time )						\
  {									\
    unsigned int tbl, tbu0, tbu1;					\
									\
    do									\
      {									\
	__asm__ __volatile__ ("mftbu %0" : "=r"(tbu0));			\
	__asm__ __volatile__ ("mftb %0" : "=r"(tbl));			\
	__asm__ __volatile__ ("mftbu %0" : "=r"(tbu1));			\
      }									\
    while (tbu0 != tbu1);						\
									\
    time = (((unsigned long long)tbu0) << 32) | tbl;			\
  }									

#define hires_timer_ui( time )	__asm__ __volatile__ ("mftb %0" : "=r"(time));


#endif /* __timers__ */
