/*
*******************************************************************************
*
* File:         main_spu.c
* RCS:          $Header: $
* Description:  The SPU driver 
* Author:       Fabrizio Petrini 
* Created:
* Modified:
* Language:     C
* Package:      N/A
* Status:       Experimental (Do Not Distribute)
*
*******************************************************************************
*/


#include <stdlib.h>
#include <assert.h>
#include <spu_intrinsics.h>
#include <spu_mfcio.h>
#include "main_spu.h"
#include "runtime_spu.h"


/* control block */
control_block_t lcb align128;
unsigned control_block_address;


/*
*******************************************************************************
*
*                               control_block
*
*******************************************************************************
*/


void get_user_control_block( void )
{
  /* receive start signal with the address of the user control block */
  control_block_address = spu_readch( SPU_RdInMbox );


  /* copying the control block in the local control block */
  blocking_get( &lcb, control_block_address, sizeof(control_block_t), 
		DMA_TAG_BLOCKING );
}


/*
*******************************************************************************
*
*                                  main
*
*******************************************************************************
*/


int main( UDB(unused) unsigned long long speid, addr64 argp, 
	  unused addr64 envp)
{
  unsigned i;
#ifdef PROFILE_SPU
  unsigned iterations, start_time, end_time;
#endif /* PROFILE_SPU */

  initialize_spu_thread( argp );

  get_user_control_block();

  barrier(DMA_TAG_BLOCKING);

  if ( my_id == 0 ) 
    {
      spe_printf("SPU: Starting SPUs:");
      spe_fflush(stdout);
    }

  barrier(DMA_TAG_BLOCKING);
  spe_printf(" %i", my_id);
  spe_fflush(stdout);

  barrier(DMA_TAG_BLOCKING);
  if ( my_id == 0 ) 
    {
      spe_printf("\n");
      spe_fflush(stdout);
    }

  barrier(DMA_TAG_BLOCKING);

#ifdef DEBUG_NFA
  if ( my_id == 0 )
    print_user_control_block();
#endif /* DEBUG_NFA */

#ifdef PROFILE_SPU
  barrier(DMA_TAG_BLOCKING);
  /* start the timer */
  spu_writech( SPU_WrDec, DECR_COUNT );
  PROF_CLEAR;
  PROF_START;
#endif /* PROFILE_SPU */

  // do something interesting here
  iterations = 27;
  start_time = 31; 
  
  printf("-----> doing something here\n" );

#ifdef PROFILE_SPU
  PROF_STOP;

  end_time = DECR_COUNT - spu_readch( SPU_RdDec );
  log_event( end_time, RUN_TIME_EVENT );

  barrier(DMA_TAG_BLOCKING);

  log_event( iterations, ITERATIONS_EVENT );

  for( i = 0; i < spus; i++ )
    {
      if ( my_id == i )
	{
	  spe_printf( "SPU %u: end time %u iterations %u\n", 
		      my_id, end_time, iterations );
	  spe_fflush(stdout);
	}
      barrier(DMA_TAG_BLOCKING);
    }

#endif /* PROFILE_SPU */

  for( i = 0; i < spus; i++ )
    {
      if ( my_id == i )
	{
	  spe_printf( "SPU %u: exiting\n", my_id );
	  spe_fflush(stdout);
	}
      barrier(DMA_TAG_BLOCKING);
    }

  dump_large_buffer();

  return 0;
}



